import sys
from functools import partial
from .inspect import get_parent_frame, get_stack_frame
from .printing import bubbletext
import time


def embed(parent_locals=None, parent_globals=None, exec_lines=None,
          remove_pyqt_hook=True, n=0):
    """
    Starts interactive session. Similar to keyboard command in matlab.
    Wrapper around IPython.embed

    """
    import IPython

    if parent_globals is None:
        parent_globals = get_parent_frame(n=n).f_globals
    if parent_locals is None:
        parent_locals = get_parent_frame(n=n).f_locals

    stackdepth = n  # NOQA
    getframe = partial(get_parent_frame, n=n)  # NOQA

    print('')
    print('================')
    print(bubbletext('EMBEDDING'))
    print('================')
    print('[util] embedding')
    try:
        if remove_pyqt_hook:
            try:
                import guitool
                guitool.remove_pyqt_input_hook()
            except (ImportError, ValueError, AttributeError) as ex:
                print('ex = {!r}'.format(ex))
                pass
            # make qt not loop forever (I had qflag loop forever with this off)
    except ImportError as ex:
        print(ex)

    # Disable common annoyance loggers
    import logging
    logging.getLogger('parso').setLevel(logging.INFO)


    #from IPython.config.loader import Config
    # cfg = Config()
    #config_dict = {}
    #if exec_lines is not None:
    #    config_dict['exec_lines'] = exec_lines
    #IPython.embed(**config_dict)
    print('[util]  Get stack location with: ')
    print('[util] get_parent_frame(n=8).f_code.co_name')
    print('[util] set EXIT_NOW or qqq to True(ish) to hard exit on unembed')
    #print('set iup to True to draw plottool stuff')
    print('[util] call %pylab qt4 to get plottool stuff working')
    once = True
    # Allow user to set iup and redo the loop
    while once or vars().get('iup', False):
        if not once:
            # SUPER HACKY WAY OF GETTING FIGURES ON THE SCREEN BETWEEN UPDATES
            #vars()['iup'] = False
            # ALL YOU NEED TO DO IS %pylab qt4
            print('re-emebeding')
            #import plottool as pt
            #pt.update()
            #(pt.present())
            for _ in range(100):
                time.sleep(.01)

        once = False
        #vars().get('iup', False):
        print('[util] calling IPython.embed()')
        """
        Notes:
            /usr/local/lib/python2.7/dist-packages/IPython/terminal/embed.py
            IPython.terminal.embed.InteractiveShellEmbed

            # instance comes from  IPython.config.configurable.SingletonConfigurable.instance
        """
        #c = IPython.Config()
        #c.InteractiveShellApp.exec_lines = [
        #    '%pylab qt4',
        #    '%gui qt4',
        #    "print 'System Ready!'",
        #]
        #IPython.embed(config=c)
        parent_ns = parent_globals.copy()
        parent_ns.update(parent_locals)
        locals().update(parent_ns)
        try:
            IPython.embed()
        except RuntimeError as ex:
            print('ex = {!r}'.format(ex))
            print('Failed to open ipython')
        #config = IPython.terminal.ipapp.load_default_config()
        #config.InteractiveShellEmbed = config.TerminalInteractiveShell
        #module = sys.modules[parent_globals['__name__']]
        #config['module'] = module
        #config['module'] = module
        #embed2(stack_depth=n + 2 + 1)
        #IPython.embed(config=config)
        #IPython.embed(config=config)
        #IPython.embed(module=module)
        # Exit python immediately if specifed
        if vars().get('EXIT_NOW', False) or vars().get('qqq', False):
            print('[jammy.embed] EXIT_NOW specified')
            sys.exit(1)


class EmbedOnException(object):
    """
    Context manager which embeds in ipython if an exception is thrown
    """
    def __init__(self):
        pass

    def __enter__(self):
        return self

    def __call__(self):
        return self

    def __exit__(__self, __type, __value, __trace):
        if __trace is not None:
            print('!!! EMBED ON EXCEPTION !!!')
            print('[util_dbg] %r in context manager!: %s ' % (__type, str(__value)))
            import traceback
            traceback.print_exception(__type, __value, __trace)
            # Grab the context of the frame where the failure occurred
            __trace_globals = __trace.tb_frame.f_globals
            __trace_locals = __trace.tb_frame.f_locals
            __trace_ns = __trace_globals.copy()
            __trace_ns.update(__trace_locals)
            # Hack to bring back names that we clobber
            if '__self' in __trace_ns:
                __self = __trace_ns['__self']
            locals().update(__trace_ns)  # I don't think this does anythign
            embed()


def fix_embed_globals():
    """
    HACK adds current locals() to globals().
    Can be dangerous.

    Solves the following issue:

        def foo():
            x = 5
            # You embed here
            import xdev
            xdev.embed()

            '''
            Now you try and run this line manually but you get a NameError

            result = [x + i for i in range(10)]

            No problem, just use. It changes all local variables to globals
            xdev.fix_embed_globals()

            '''
            result = [x + i for i in range(10)]

        foo()
    """
    # Get the stack frame of whoever called this function
    frame = get_stack_frame(n=1)
    # Hack all of the local variables to be global variables
    frame.f_globals.update(frame.f_locals)
    # Leave some trace that we did this
    frame.f_globals['_did_xdev_fix_embed_globals'] = True


embed_on_exception_context = EmbedOnException()


import gpustat
import numpy as np


def select_gpu(mem_prior=1.0):
    mem_prior = np.clip(mem_prior, 0.0,1.0)
    query = gpustat.new_query()
    if len(query) == 0:
        raise RuntimeError("gpu not available")
    if len(query) == 1:
        return query[0].entry["index"]

    mem_list, utils_list = get_mem_util()
    mem, utils = np.array(mem_list), np.array(utils_list)
    weight = mem*mem_prior + utils * (1-mem_prior)
    least_utils_id = np.argmin(weight)

    return query[least_utils_id].entry["index"]


def gpu_by_util():
    query = gpustat.new_query()
    _, utils_list = get_mem_util()
    ids = np.argsort(utils_list)
    return [query[id_item].entry["index"] for id_item in ids]


def get_mem_util():
    query = gpustat.new_query()
    used_space_list = [1.0 * item.memory_used/item.memory_total for item in query]
    utils_list = [item.utilization for item in query]
    return used_space_list, utils_list



from subprocess import PIPE, Popen

__all__ = ["run_simple_command"]


def run_simple_command(cmd: str):
    cmd = cmd.split(" ")
    with Popen(cmd, stdout=PIPE, stderr=PIPE) as prc:
        stdout, stderr = prc.communicate()
        stdout = stdout.decode("utf-8")
        stderr = stderr.decode("utf-8")
        prc.terminate()
    return stdout, stderr



import matplotlib.pyplot as plt

__all__ = ["jcolors"]

# jcolors = ["blue", "orange", "green", "red", "purple", "cyan", "brown"]
# https://stackoverflow.com/questions/42086276/get-default-line-colour-cycle
jcolors = plt.rcParams["axes.prop_cycle"].by_key()["color"]



import os
import time
import uuid

import gpustat

from jammy.comm import get_local_addr, is_port_used
from jammy.comm.cs import ClientPipe, ServerPipe
from jammy.logging import get_logger
from jammy.utils import gpu
from jammy.utils.env import jam_getenv
from jammy.utils.printing import kvformat

logger = get_logger()


def req_util(pipe, identifier, inp=None):
    num_gpus = inp.get("num_gpus", 1)
    sleep_sec = inp.get("sleep_sec", 3)
    idx = identifier.decode("ascii")
    logger.info(f"REQ from {idx}\n {kvformat(inp)}")
    gpu_list = gpu.gpu_by_util()
    msg = None
    if len(gpu_list) < num_gpus:
        msg = gpu_list
    else:
        msg = gpu_list[:num_gpus]
    pipe.send(identifier, msg)
    logger.info(f"SEND: {idx} {msg}")
    time.sleep(sleep_sec)
    logger.info(f"CLOSE: {idx}")


def len_available_gpu(pipe, identifier, inp=None):
    del inp
    pipe.send(identifier, len(gpustat.new_query()))


def start_sever():
    server = ServerPipe("server", mode="tcp")
    server.dispatcher.register("req_util", req_util)
    server.dispatcher.register("total", len_available_gpu)
    p_router = jam_getenv("GPUPort", default=1080, type=int)
    assert not is_port_used(p_router)
    assert not is_port_used(p_router + 1)
    with server.activate(tcp_port=[str(p_router), str(p_router + 1)]):
        logger.critical("GPU ROUTER and PULL IP:")
        logger.critical(server.conn_info)
        while True:
            time.sleep(1)


def instantiate_client(flag="_"):
    p_dealer = jam_getenv("GPUPort", default=1080, type=int)
    tcp_dealer = f"tcp://{get_local_addr()}:{p_dealer}"
    tcp_push = f"tcp://{get_local_addr()}:{p_dealer+1}"
    client = ClientPipe(
        f"gpu_client{flag}" + uuid.uuid4().hex[:8], conn_info=[tcp_dealer, tcp_push]
    )
    return client


def start_client():
    client = instantiate_client()
    logger.info("Identity: {}.".format(client.identity))
    with client.activate():
        echo = client.query("total")
        logger.info(f"Contains {echo} gpus")
        query_gpu_ids = client.query("req_util", dict(num_gpus=1))
        logger.info(f"req get {query_gpu_ids}")


def get_gpu_by_utils(num_gpus: int = 1, sleep_sec: int = 3):
    client = instantiate_client(os.getpid())
    with client.activate():
        inp = {
            "num_gpus": num_gpus,
            "sleep_sec": sleep_sec,
        }
        query_gpu_ids = client.query("req_util", inp)
    return query_gpu_ids



import os

import matplotlib.pyplot as plt
import numpy as np
import torch as th
from einops import rearrange
from torchvision.utils import make_grid, save_image

__all__ = [
    "show_imgs",
    "show_imgs_traj",
    "save_seperate_imgs",
]


def show_imgs(img_data, nrow=10):
    grid_img = make_grid(img_data, nrow)
    plt.figure(figsize=(nrow, 8))
    plt.axis("off")
    plt.imshow(grid_img.permute(1, 2, 0).cpu())


def show_imgs_traj(traj, num_img, num_steps):
    idx = np.linspace(0, len(traj) - 1, num_steps, dtype=int)
    imgs = []
    for cur_idx in idx:
        imgs.append(traj[cur_idx][:num_img])
    imgs = th.cat(imgs)

    imgs = rearrange(imgs, "(n b) ... -> (b n) ...", b=num_img)
    show_imgs(imgs, num_steps)


def save_seperate_imgs(sample, sample_path, cnt):
    batch_size = len(sample)
    for i in range(batch_size):
        save_image((sample[i] + 1.0) / 2, os.path.join(sample_path, f"{cnt:07d}.png"))
        cnt += 1



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jammy',
 'jammy.cli',
 'jammy.collections',
 'jammy.comm',
 'jammy.concurrency',
 'jammy.event',
 'jammy.image',
 'jammy.io',
 'jammy.logging',
 'jammy.random',
 'jammy.storage',
 'jammy.storage.kv',
 'jammy.utils',
 'jamnp',
 'jamtorch',
 'jamtorch.cuda',
 'jamtorch.data',
 'jamtorch.ddp',
 'jamtorch.distributions',
 'jamtorch.io',
 'jamtorch.learn',
 'jamtorch.learn.gan',
 'jamtorch.logging',
 'jamtorch.nn',
 'jamtorch.nn.utils',
 'jamtorch.pl_callbacks',
 'jamtorch.prototype',
 'jamtorch.trainer',
 'jamtorch.utils',
 'jamviz',
 'jamviz.jupyter',
 'jamviz.plt',
 'jamweb',
 'jamweb.session',
 'jamweb.web']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.17,<4.0.0',
 'attrs>=20.3.0,<21.0.0',
 'einops>=0.3.0,<0.4.0',
 'gpustat>=0.6.0,<0.7.0',
 'hydra-core>=1.1.0,<2.0.0',
 'ipdb>=0.13.8,<0.14.0',
 'loguru>=0.5.3,<0.6.0',
 'matplotlib>=3.3.0,<4.0.0',
 'numpy>=1.20.3,<2.0.0',
 'scipy>=1.6.3,<2.0.0',
 'tqdm>=4.11.0,<5.0.0']

extras_require = \
{'all': ['wandb>=0.10.31,<0.11.0',
         'pytorch-lightning>=1.5.1,<2.0.0',
         'torch>=1.9.0,<2.0.0',
         'torchvision>=0.10.0,<0.11.0',
         'torchinfo>=1.5.3,<2.0.0',
         'tornado>=6.1,<7.0',
         'pyzmq>=22.3.0,<23.0.0',
         'h5py>=3.4.0,<4.0.0',
         'msgpack>=1.0.2,<2.0.0',
         'msgpack-numpy>=0.4.7,<0.5.0',
         'pyarrow>=6.0.0,<7.0.0',
         'lmdb>=1.2.1,<2.0.0',
         'python-memcached>=1.59,<2.0',
         'line-profiler>=3.3.1,<4.0.0',
         'ubelt>=0.10.2,<0.11.0',
         'parse>=1.19.0,<2.0.0',
         'pudb>=2022.1,<2023.0'],
 'pro': ['line-profiler>=3.3.1,<4.0.0',
         'ubelt>=0.10.2,<0.11.0',
         'parse>=1.19.0,<2.0.0',
         'pudb>=2022.1,<2023.0'],
 'storage': ['h5py>=3.4.0,<4.0.0',
             'msgpack>=1.0.2,<2.0.0',
             'msgpack-numpy>=0.4.7,<0.5.0',
             'pyarrow>=6.0.0,<7.0.0',
             'lmdb>=1.2.1,<2.0.0',
             'python-memcached>=1.59,<2.0'],
 'torch': ['wandb>=0.10.31,<0.11.0',
           'pytorch-lightning>=1.5.1,<2.0.0',
           'torch>=1.9.0,<2.0.0',
           'torchvision>=0.10.0,<0.11.0',
           'torchinfo>=1.5.3,<2.0.0'],
 'web': ['tornado>=6.1,<7.0', 'pyzmq>=22.3.0,<23.0.0']}

entry_points = \
{'console_scripts': ['jgpuc = jammy.cli.gpu_sc:start_client',
                     'jgpus = jammy.cli.gpu_sc:start_sever']}

setup_kwargs = {
    'name': 'jammy',
    'version': '0.0.7a3',
    'description': 'Personal ToolBox',
    'long_description': '<h1 align="center"> Jammy (Jam) </h1>\n\n<p align="center">\n  <a href="https://pypi.org/project/jammy/">\n    <img src="https://img.shields.io/pypi/v/jammy?style=for-the-badge" alt="PyPI" />\n  </a>\n  <a href="#">\n    <img src="https://img.shields.io/pypi/l/jammy?style=for-the-badge" alt="PyPI - License" />\n  </a>\n  <a href="https://github.com/qsh-zh/jam">\n    <img src="https://img.shields.io/badge/-github-grey?style=for-the-badge&logo=github" alt="GitHub code" />\n  </a>\n  <a href="https://gitlab.com/qsh.zh/jam">\n    <img src="https://img.shields.io/badge/-gitlab-grey?style=for-the-badge&logo=gitlab" alt="GitLab code" />\n  </a>\n  <a href="https://jammy.readthedocs.io/en/stable/index.html">\n    <img src="https://img.shields.io/readthedocs/jammy?style=for-the-badge" alt="Read the Docs" />\n  </a>\n  <a href="#">\n    <img src="https://img.shields.io/pypi/pyversions/jammy?style=for-the-badge" alt="PyPI - Python Version" />\n  </a>\n  <a href="https://github.com/psf/black">\n    <img src="https://img.shields.io/badge/code%20style-black-000000.svg?style=for-the-badge" alt="Code style: black" />\n  </a>\n  <p align="center">\n    <i>A personal toolbox by <a href="https://qsh-zh.github.io/">Qsh.zh</a>.</i>\n  </p>\n</p>\n\n### Usage\n\n#### setup\n\n* For core package, run `pip install jammy`\n* To access functions in `bin`\n```shell\ngit clone https://gitlab.com/qsh.zh/jam.git --recursive\nexport PATH=<path_to_jam>/bin:$PATH\n# run python program\njam-run main.py\njam-crun 1 main.py # use second nvidia gpu\n```\n\n\n#### sample of io\n```python\nimport jammy.io as jio\nfrom jamtorch.utils import as_numpy\njio.dump("ndarray.npz", np.arange(10))\njio.dump("foo.pkl", {"foo": as_numpy(torch.arange(10).cuda())})\nndarray = jio.load("ndarray.npz")\njio.load("foo.pkl")\nmodel_dict = jio.load("checkpoint.pth")\n```\n```shell\n$ jam-inspect-file foo.pkl\n> python3\n[ins] print(f1)\n# content of foo.pkl\n\n```\n\n### Advanced Usage\n\n* [A DDP pytorch training framework](https://jammy.readthedocs.io/en/stable/jamtorch.ddp.html)\n* [Registry](https://jammy.readthedocs.io/en/stable/jammy.utils.html?highlight=registry#jammy.utils.registry.CallbackRegistry)\n* TODO\n\n### Etymology\n* The naming is inspired from [Jyutping](https://en.wikipedia.org/wiki/Jyutping) of [Qin](https://en.wiktionary.org/wiki/%E6%AC%BD).\n\n### MICS\n\n* The package and framework are inspired from [Jacinle](https://github.com/vacancy/Jacinle) by [vacancy](https://github.com/vacancy), from which I learn and take utility functions shamelessly.\n',
    'author': 'Qin',
    'author_email': 'qsh.zh27@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)
